/**
 ****************************************************************************************
 *
 * @file ch_map.c
 *
 * @brief Channel map display function
 *
 * Copyright (C) 2018 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#if dg_configUSE_CH_MAP

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "sdk_defs.h"
#include <stdbool.h>
#include "FreeRTOS.h"
#include "osal.h"
#include "timers.h"
#include "co_bt.h"
#include "llc.h"

#define CH_MAP_PT_REPORT_TIME 200

#if (dg_configUSE_LP_CLK == LP_CLK_32000)
#define SOCF_REF_LP_FACTOR ((1000 << 20) / 32000)
#elif (dg_configUSE_LP_CLK == LP_CLK_32768)
#define SOCF_REF_LP_FACTOR ((1000 << 20) / 32768)
#elif (dg_configUSE_LP_CLK == LP_CLK_RCX)
#define SOCF_REF_LP_FACTOR ((1000 << 20) / rcx_clock_hz)
#elif (dg_configUSE_LP_CLK == LP_CLK_ANY)
// Must be defined in the custom_config_<>.h file.
#else
# error "CH_MAP_REF_LP_FACTOR is not defined!"
#endif

void int2bin(int8_t n, char s[])
{
    // determine the number of bits needed ("sizeof" returns bytes)
    int8_t nbits = sizeof(n) * 8;
    // char *s = OS_MALLOC(nbits+1);  // +1 for '\0' terminator
    s[nbits] = '\0';
    // forcing evaluation as an unsigned value prevents complications
    // with negative numbers at the left-most bit
    uint8_t u = *(uint8_t *)&n;
    int8_t i;
    uint8_t mask = 1 << (nbits-1); // fill in values right-to-left
    for (i = 0; i < nbits; i++, mask >>= 1)
        s[i] = ((u & mask) != 0) + '0';
    //OS_FREE(s);
    return;
}

uint8_t reverseBits(uint8_t num)
{
    uint8_t  NO_OF_BITS = sizeof(num) * 8;
    uint8_t reverse_num = 0;
    uint8_t i;
    for (i = 0; i < NO_OF_BITS; i++)
    {
        if((num & (1 << i)))
           reverse_num |= 1 << ((NO_OF_BITS - 1) - i);
   }
    return reverse_num;
}



#if defined(CONFIG_RETARGET) && dg_configUSE_CH_MAP
#define CH_MAP_NOTIF              (1 << 16)
#define ch_map_TASK_PRIORITY      ( tskIDLE_PRIORITY + 1 )
//__RETAINED static int32_t socf_log_time = 0;
PRIVILEGED_DATA static OS_TASK xSocfTaskHandle;
PRIVILEGED_DATA static TimerHandle_t xch_mapReport;

static void ch_map_report_cb(TimerHandle_t timer)
{
        OS_TASK task = (OS_TASK)pvTimerGetTimerID(timer);
        OS_TASK_NOTIFY(task, CH_MAP_NOTIF, eSetBits);
}

/**
 * \brief Channel_Map Task function.
 *
 * \param [in] pvParameters ignored.
 *
 */
static void ch_map_task(void *pvParameters)
{
        uint32_t ulNotifiedValue;
        uint8_t i;
        volatile struct le_chnl_map channel_map;
        volatile uint8_t c_map[5];
        char bin_map[5][9];

        OS_BASE_TYPE xResult __attribute__((unused));

        xch_mapReport = OS_TIMER_CREATE("CH_MAP_Report",
                                OS_MS_2_TICKS(CH_MAP_PT_REPORT_TIME),
                                OS_TIMER_FAIL,
                                (void *) OS_GET_CURRENT_TASK(),
                                ch_map_report_cb);
        OS_ASSERT(xch_mapReport != NULL);

        OS_TIMER_START(xch_mapReport, OS_TIMER_FOREVER);

        for(;;) {
                xResult = OS_TASK_NOTIFY_WAIT(0x0, 0xFFFFFFFF, &ulNotifiedValue,
                        portMAX_DELAY);

                if (ulNotifiedValue & CH_MAP_NOTIF) {
                      channel_map = llc_env[0]->ch_map;
                       for (i=0; i<5; i++)
                       {
                    	   c_map[i] = reverseBits(channel_map.map[(4-i)]);
                    	   int2bin (c_map[i], bin_map[i]);
                       }
                       printf("Channel Map: %8s%8s%8s%8s%8s\r\n", bin_map[4], bin_map[3], bin_map[2], bin_map[1], bin_map[0] );
                       OS_TIMER_START(xch_mapReport, OS_TIMER_FOREVER);
                }
        }
}

void ch_map_report_init(void)
{
        OS_BASE_TYPE status;

        status = OS_TASK_CREATE("CH_MAP",
                                ch_map_task,
                                ( void * ) NULL,
                                768,
                                ch_map_TASK_PRIORITY,
                                xSocfTaskHandle);
        OS_ASSERT(status == OS_OK);

        (void)status;
}
#endif

#endif /* dg_configUSE_CH_MAP */
